# Copyright (c) Glow Contributors. See CONTRIBUTORS file.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This was mostly taken from a tutorial from Caffe2:
# caffe2/blob/master/caffe2/python/tutorials/py_gen/MNIST.py

# It currently allows to train either LeNet or an MLP on MNIST. We can then load
# the pre-trained protobuf file into Glow to run.

import numpy as np
import os
import shutil
import caffe2.python.predictor.predictor_exporter as pe

from caffe2.python.predictor import mobile_exporter

from caffe2.python import brew, core, model_helper, optimizer, workspace

# If you would like to see some really detailed initializations,
# you can change --caffe2_log_level=0 to --caffe2_log_level=-1
core.GlobalInit(["caffe2", "--caffe2_log_level=0"])
print("Necessities imported!")

# If True, a more complicated convolutional model is used
# If False, a multilayer perceptron model is used
USE_LENET_MODEL = True

# This section preps your image and test set in a lmdb database


def DownloadResource(url, path):
    """Downloads resources from s3 by url and unzips them to the provided path"""
    import requests
    import zipfile
    import StringIO

    print("Downloading... {} to {}".format(url, path))
    r = requests.get(url, stream=True)
    z = zipfile.ZipFile(StringIO.StringIO(r.content))
    z.extractall(path)
    print("Completed download and extraction.")


current_folder = os.path.join(os.path.expanduser("~"), "caffe2_notebooks")
data_folder = os.path.join(current_folder, "tutorial_data", "mnist")
root_folder = os.path.join(current_folder, "tutorial_files", "tutorial_mnist")
db_missing = False

if not os.path.exists(data_folder):
    os.makedirs(data_folder)
    print("Your data folder was not found!! This was generated: {}".format(data_folder))

# Look for existing database: lmdb
# MNIST lmdb can be found here:
# https://download.caffe2.ai/databases/mnist-lmdb.zip
if os.path.exists(os.path.join(data_folder, "mnist-train-nchw-lmdb")):
    print("lmdb train db found!")
else:
    db_missing = True

if os.path.exists(os.path.join(data_folder, "mnist-test-nchw-lmdb")):
    print("lmdb test db found!")
else:
    db_missing = True

# attempt the download of the db if either was missing
if db_missing:
    print("one or both of the MNIST lmbd dbs not found!!")
    db_url = "http://download.caffe2.ai/databases/mnist-lmdb.zip"
    try:
        DownloadResource(db_url, data_folder)
    except Exception as ex:
        print(
            "Failed to download dataset. Please download it manually from {}".format(
                db_url
            )
        )
        print(
            "Unzip it and place the two database folders here: {}".format(
                data_folder)
        )
        raise ex

if os.path.exists(root_folder):
    print("Looks like you ran this before, so we need to cleanup those old files...")
    shutil.rmtree(root_folder)

os.makedirs(root_folder)
workspace.ResetWorkspace(root_folder)

print("training data folder:" + data_folder)
print("workspace root folder:" + root_folder)


def AddInput(model, batch_size, db, db_type):
    # load the data
    data_uint8, label = brew.db_input(
        model,
        blobs_out=["data_uint8", "label"],
        batch_size=batch_size,
        db=db,
        db_type=db_type,
    )

    print(data_uint8._from_net)

    # cast the data to float
    data = model.Cast(data_uint8, "data", to=core.DataType.FLOAT)
    # scale data from [0,255] down to [0,1]
    data = model.Scale(data, data, scale=float(1.0 / 256))
    # don't need the gradient for the backward pass
    data = model.StopGradient(data, data)
    return data, label


def AddMLPModel(model, data):
    size = 28 * 28 * 1
    sizes = [size, size * 2, size * 2, 10]
    layer = data
    for i in range(len(sizes) - 1):
        layer = brew.fc(
            model,
            layer,
            "dense_{}".format(i),
            dim_in=sizes[i],
            dim_out=sizes[i + 1],
            use_cudnn=False,
        )
        layer = model.net.Relu(layer, "relu_{}".format(i), use_cudnn=False)
    softmax = model.net.Softmax(layer, "softmax", use_cudnn=False)
    return softmax


def AddLeNetModel(model, data):
    """
    This part is the standard LeNet model: from data to the softmax prediction.

    For each convolutional layer we specify dim_in - number of input channels
    and dim_out - number or output channels. Also each Conv and MaxPool layer changes the
    image size. For example, kernel of size 5 reduces each side of an image by 4.

    While when we have kernel and stride sizes equal 2 in a MaxPool layer, it divides
    each side in half.
    """
    # Image size: 28 x 28 -> 24 x 24
    conv1 = brew.conv(
        model, data, "conv1", dim_in=1, dim_out=20, kernel=5, use_cudnn=False
    )
    # Image size: 24 x 24 -> 12 x 12
    pool1 = model.net.MaxPool(
        conv1,
        "pool1",
        kernel=2,
        stride=2,
        use_cudnn=False)
    # Image size: 12 x 12 -> 8 x 8
    conv2 = brew.conv(
        model, pool1, "conv2", dim_in=20, dim_out=50, kernel=5, use_cudnn=False
    )
    # Image size: 8 x 8 -> 4 x 4
    pool2 = model.net.MaxPool(
        conv2,
        "pool2",
        kernel=2,
        stride=2,
        use_cudnn=False)
    # 50 * 4 * 4 stands for dim_out from previous layer multiplied by the
    # image size
    fc3 = brew.fc(
        model,
        pool2,
        "fc3",
        dim_in=50 * 4 * 4,
        dim_out=500,
        use_cudnn=False)
    fc3 = model.net.Relu(fc3, "relu3", use_cudnn=False)
    pred = brew.fc(model, fc3, "pred", 500, 10, use_cudnn=False)
    softmax = model.net.Softmax(pred, "softmax", use_cudnn=False)
    return softmax


def AddModel(model, data):
    if USE_LENET_MODEL:
        return AddLeNetModel(model, data)
    else:
        return AddMLPModel(model, data)


def AddAccuracy(model, softmax, label):
    """Adds an accuracy op to the model"""
    accuracy = model.Accuracy([softmax, label], "accuracy", use_cudnn=False)
    return accuracy


def AddTrainingOperators(model, softmax, label):
    """Adds training operators to the model."""
    xent = model.LabelCrossEntropy([softmax, label], "xent", use_cudnn=False)
    # compute the expected loss
    loss = model.AveragedLoss(xent, "loss", use_cudnn=False)
    # track the accuracy of the model
    AddAccuracy(model, softmax, label)
    # use the average loss we just computed to add gradient operators to the
    # model
    model.AddGradientOperators([loss])
    optimizer.build_sgd(
        model, base_learning_rate=0.1, policy="step", stepsize=1, gamma=0.999
    )


arg_scope = {"order": "NCHW"}
train_model = model_helper.ModelHelper(name="mnist_train", arg_scope=arg_scope)
data, label = AddInput(
    train_model,
    batch_size=64,
    db=os.path.join(data_folder, "mnist-train-nchw-lmdb"),
    db_type="lmdb",
)
softmax = AddModel(train_model, data)
AddTrainingOperators(train_model, softmax, label)

test_model = model_helper.ModelHelper(
    name="mnist_test", arg_scope=arg_scope, init_params=False
)
data, label = AddInput(
    test_model,
    batch_size=100,
    db=os.path.join(data_folder, "mnist-test-nchw-lmdb"),
    db_type="lmdb",
)
softmax = AddModel(test_model, data)

# Deployment model. We simply need the main AddModel part.
deploy_model = model_helper.ModelHelper(
    name="mnist_deploy", arg_scope=arg_scope, init_params=False
)
AddModel(deploy_model, "data")

# The parameter initialization network only needs to be run once.
# Now all the parameter blobs are going to be initialized in the workspace.
workspace.RunNetOnce(train_model.param_init_net)

# overwrite=True allows you to run this cell several times and avoid errors
workspace.CreateNet(train_model.net, overwrite=True)

# Set the iterations number and track the accuracy & loss
total_iters = 200
accuracy = np.zeros(total_iters)
loss = np.zeros(total_iters)

print("The blobs in the workspace pre-train: {}".format(workspace.Blobs()))

# Now, we will manually run the network for 200 iterations.
for i in range(total_iters):
    workspace.RunNet(train_model.net)
    accuracy[i] = workspace.blobs["accuracy"]
    loss[i] = workspace.blobs["loss"]

print("The blobs in the workspace post-train: {}".format(workspace.Blobs()))


# param_init_net here will only create a data reader
# Other parameters won't be re-created because we selected
# init_params=False before
workspace.RunNetOnce(test_model.param_init_net)
workspace.CreateNet(test_model.net, overwrite=True)
test_accuracy = np.zeros(100)
for i in range(100):
    workspace.RunNet(test_model.net.Proto().name)
    test_accuracy[i] = workspace.FetchBlob("accuracy")

print("test_accuracy: %f" % test_accuracy.mean())

# construct the model to be exported
# the inputs/outputs of the model are manually specified.
pe_meta = pe.PredictorExportMeta(
    predict_net=deploy_model.net.Proto(),
    parameters=[str(b) for b in deploy_model.params],
    inputs=["data"],
    outputs=["softmax"],
)

# save the model to a file. Use minidb as the file format
pe.save_to_db(
    "minidb",
    os.path.join(
        root_folder,
        "mnist_model.minidb"),
    pe_meta)
print("The deploy model is saved to: " + root_folder + "/mnist_model.minidb")

workspace.RunNetOnce(deploy_model.param_init_net)

init_net, predict_net = mobile_exporter.Export(
    workspace, deploy_model.net, deploy_model.params
)
with open("init_net.pb", "wb") as f:
    f.write(init_net.SerializeToString())
with open("predict_net.pb", "wb") as f:
    f.write(predict_net.SerializeToString())
with open("predict_net.pbtxt", "wb") as f:
    f.write(str(deploy_model.net.Proto()))

# Now we can load the model back and run the prediction to verify it works.

# we retrieve the last input data out and use it in our prediction test
# before we scratch the workspace
blob = workspace.FetchBlob("data")

# reset the workspace, to make sure the model is actually loaded
workspace.ResetWorkspace(root_folder)

# verify that all blobs are destroyed.
print("The blobs in the workspace after reset: {}".format(workspace.Blobs()))

# load the predict net
predict_net = pe.prepare_prediction_net(
    os.path.join(root_folder, "mnist_model.minidb"), "minidb"
)

# verify that blobs are loaded back
print(
    "The blobs in the workspace after loading the model: {}".format(
        workspace.Blobs())
)

# feed the previously saved data to the loaded model
workspace.FeedBlob("data", blob)

# predict
workspace.RunNetOnce(predict_net)
softmax = workspace.FetchBlob("softmax")
